#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdarg.h>
#include <png.h>
#include <errno.h>
#include "pngio.h"

#define SIG_BYTES 8


static char *prepare_filep_for_read(FILE *fp,const char *fni)
{
  char ebu[2000];
  unsigned char header[SIG_BYTES];

  if (fread(header, 1, SIG_BYTES, fp)!=SIG_BYTES)
  {
    snprintf(ebu,sizeof(ebu),"can not read signature bytes from file %s", fni);
    return strdup(ebu);
  }

  if (png_sig_cmp (header, 0, SIG_BYTES))
  {
    snprintf(ebu,sizeof(ebu),"file %s is not a PNG file.", fni);
    return strdup(ebu);
  }
  return NULL;
}

static char *open_png_file(const char *fn, FILE **fp)
{
  FILE *f;
  char *errstr;
  char ebu[2000];

  f = fopen (fn, "rb");
  if (!f)
  {   
    snprintf(ebu,sizeof(ebu),"can not open file %s: %s", fn, strerror (errno));
    return strdup(ebu);
  }
  errstr= prepare_filep_for_read(f, fn);
  if (errstr)
    fclose(f);
  else
    *fp= f;
  return errstr;
}


static char *do_read_png
  (png_structp png_ptr,
   unsigned int *Rw, unsigned int *Rh, unsigned int **Rdata)
{
  char *errstr;
  png_infop info_ptr;
  png_infop end_info;
  unsigned int *data, **rows;
  png_uint_32 ww, hh, i;
  int bit_depth, color_type, interlace_type;

  info_ptr= NULL;
  end_info= NULL;
  rows= NULL;
  data= NULL;

  info_ptr = png_create_info_struct (png_ptr);
  if (!info_ptr)
  {
    errstr= strdup("internal error. libpng: "
                    "can not allocate first info struct.");
    goto done;
  }

  end_info = png_create_info_struct (png_ptr);
  if (!end_info)
  {
    errstr= strdup("internal error. libpng: "
                    "can not allocate second info struct.");
    goto done;
  }

  if (setjmp (png_jmpbuf (png_ptr)))
  {
    errstr= strdup("internal error. libpng jumped back.");
    goto done;
  }

  // Get all of the information from the PNG file before the first IDAT (imaged ata chunk).
  png_read_info (png_ptr, info_ptr);

  // Get the goods
  png_get_IHDR (png_ptr, info_ptr, &ww, &hh, &bit_depth, &color_type,
		&interlace_type, NULL, NULL);

  // tell libpng to strip 16 bit/color files down to 8 bits/color
  png_set_strip_16 (png_ptr);

  // Expand paletted colors into true RGB triplets
  if (color_type == PNG_COLOR_TYPE_PALETTE)
    png_set_expand (png_ptr);

  // Expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel
  if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
    png_set_expand (png_ptr);

  // Expand paletted or RGB images with transparency to full alpha channels
  // so the data will be available as RGBA quartets.
  if (png_get_valid (png_ptr, info_ptr, PNG_INFO_tRNS))
    png_set_expand (png_ptr);

  // Grey images are upgraded to RGB
  if (color_type == PNG_COLOR_TYPE_GRAY
      || color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
    png_set_gray_to_rgb (png_ptr);

  // If we dont have true alpha, pad with an alpha channel representing opaque
  png_set_filler (png_ptr, 0xff, PNG_FILLER_AFTER);

  // Turn on interlace handling
  png_set_interlace_handling (png_ptr);

  // Update image info based on transformations
  png_read_update_info (png_ptr, info_ptr);

  data = malloc (ww * hh * sizeof (unsigned int));
  rows = malloc (hh * sizeof (unsigned int *));
  for (i = 0; i < hh; i++)
    rows[i] = data + i * ww;

  png_read_image (png_ptr, (void *) rows);

  // read rest of file, and get additional chunks in info_ptr
  png_read_end (png_ptr, info_ptr);

  // clean up after the read, and free any memory allocated
  png_destroy_read_struct (&png_ptr, &info_ptr, &end_info);

  /* Now we need to convert from the PNG RGBA format to our uint32_t
     format. The PNG library stores the R byte first, and then G etc.
     We want A at the top 8 bits, then the R, etc. */
  unsigned char *bdata;
  bdata= (unsigned char*) data;
  for(i=0;i<ww*hh;i++)
  {
    unsigned char R,G,B,A; unsigned int P;
    R= bdata[i*4];
    G= bdata[i*4+1];
    B= bdata[i*4+2];
    A= bdata[i*4+3];
    P= (A<<24) | (R<<16) | (G<<8) | B;
    data[i]= P;
  }

  free(rows);

  *Rw = ww;
  *Rh = hh;
  *Rdata = data;

  return NULL;
done:
  if (data) free(data);
  if (rows) free(rows);
  png_destroy_read_struct (&png_ptr, info_ptr ? &info_ptr : NULL,
                                     end_info ? &end_info : NULL);
  return errstr;
}

char *read_png_image
  (const char *fn, unsigned int *Rw, unsigned int *Rh, unsigned int **Rdata)
{
  char *errstr;
  FILE *fp;
  png_structp png_ptr;

  if ((errstr=open_png_file(fn,&fp))) return errstr;

  png_ptr= png_create_read_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (!png_ptr)
  {
    errstr= strdup ("internal error. libpng: can not allocate PNG struct.");
      goto done;
  }

  png_init_io (png_ptr, fp);
  png_set_sig_bytes (png_ptr, SIG_BYTES);

  errstr= do_read_png(png_ptr, Rw, Rh, Rdata);
done:
  fclose(fp);
  return errstr;
}

typedef struct { const unsigned char *start, *end; } pbuf_t;

static void read_buffer(png_structp png_ptr, png_bytep data, png_size_t length)
{
  pbuf_t *B;
  B= png_get_io_ptr(png_ptr);
  if (B->start+length>B->end)
    png_error(png_ptr,"Requested read beyond the PNG buffer");
  memcpy(data, B->start, length);
  B->start+= length;
}

char *decode_png_image
  (const unsigned char *data, size_t len,
   unsigned int *Rw, unsigned int *Rh, unsigned int **Rdata)
{
  pbuf_t buffer;
  png_structp png_ptr;

  if (png_sig_cmp (data, 0, SIG_BYTES))
    return strdup("Not a valid PNG buffer.");

  png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);

  if (!png_ptr)
    return strdup("internal error. libpng: can not allocate PNG struct.");

  buffer.start= data+SIG_BYTES; buffer.end= data+len;
  png_set_read_fn(png_ptr, &buffer, read_buffer);
  png_set_sig_bytes (png_ptr, SIG_BYTES);

  return do_read_png(png_ptr, Rw, Rh, Rdata);
}


char *write_png_image
  (const char *fn, unsigned int w, unsigned int h, unsigned int *data)
{
  char *errstr;
  FILE *fp;
  char ebu[2000];
  png_structp png_ptr;
  png_infop info_ptr;
  uint8_t **rows;
  png_uint_32 i;
  uint8_t *tdata;

  rows= NULL; tdata= NULL;

  errstr = 0;
  fp = fopen (fn, "wb");
  if (!fp)
  {
    snprintf(ebu,sizeof(ebu),"can not open file %s: %s", fn, strerror (errno));
    return strdup(ebu);
  }

  png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (!png_ptr)
  {
    errstr= strdup ("internal error. libpng: can not allocate PNG struct.");
    goto err_ret_close;
  }

  info_ptr = png_create_info_struct (png_ptr);
  if (!info_ptr)
  {
    errstr= strdup ("internal error. libpng: can not allocate info struct.");
    goto err_ret_destroy0;
  }

  if (setjmp (png_jmpbuf (png_ptr)))
  {
    errstr = strdup ("internal error. libpng jumped back.");
    goto err_ret_destroy1;
  }

  png_init_io (png_ptr, fp);

  png_set_IHDR (png_ptr, info_ptr,
		w, h, 8,
		PNG_COLOR_TYPE_RGBA,
		PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

  png_write_info (png_ptr, info_ptr);

  rows = malloc (h * sizeof (unsigned int *));
  tdata= malloc(sizeof(uint8_t)*w*h*4);
  for(i=0;i<w*h;i++)
  {
    tdata[i*4+0]= (data[i] >> 16);   // R
    tdata[i*4+1]= (data[i] >> 8);    // G
    tdata[i*4+2]= (data[i] >> 0);    // B
    tdata[i*4+3]= (data[i] >> 24);   // A
  }
  for (i = 0; i < h; i++)
  {
    rows[i] = tdata + i*w*4;
  }

  png_write_image (png_ptr, (void *) rows);
  png_write_end (png_ptr, info_ptr);

  png_destroy_write_struct (&png_ptr, &info_ptr);
  free(tdata);
  free (rows);
  fclose (fp);

  return 0;

err_ret_destroy0:
  png_destroy_write_struct (&png_ptr, 0);
  fclose (fp);
  return errstr;
err_ret_destroy1:
  if (tdata) free(tdata);
  if (rows) free(rows);
  png_destroy_write_struct (&png_ptr, &info_ptr);
  fclose (fp);
  return errstr;
err_ret_close:
  fclose (fp);
  return errstr;
}

char *get_png_dimensions (const char *fn, unsigned int *Rw, unsigned int *Rh)
{
  char *errstr;
  FILE *fp;
  png_structp png_ptr;
  png_infop info_ptr;
  png_infop end_info;
  png_uint_32 ww, hh;
  int bit_depth, color_type, interlace_type;

  if ((errstr=open_png_file(fn, &fp))) return errstr;

  png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);

//(png_voidp)user_error_ptr, user_error_fn, user_warning_fn);

  if (!png_ptr)
    {
      errstr =
	strdup ("internal error. libpng: can not allocate PNG struct.");
      goto err_ret_close;
    }

  info_ptr = png_create_info_struct (png_ptr);
  if (!info_ptr)
    {
      errstr =
	strdup
	("internal error. libpng: can not allocate first info struct.");
      goto err_ret_destroy0;
    }

  end_info = png_create_info_struct (png_ptr);
  if (!end_info)
    {
      errstr =
	strdup
	("internal error. libpng: can not allocate second info struct.");
      goto err_ret_destroy1;
    }

  if (setjmp (png_jmpbuf (png_ptr)))
    {
      errstr = strdup ("internal error. libpng jumped back.");
      goto err_ret_destroy2;
    }

  png_init_io (png_ptr, fp);
  png_set_sig_bytes (png_ptr, SIG_BYTES);

  // Get all of the information from the PNG file before the first IDAT (imaged ata chunk).
  png_read_info (png_ptr, info_ptr);

  // Get the goods
  png_get_IHDR (png_ptr, info_ptr, &ww, &hh, &bit_depth, &color_type,
		&interlace_type, NULL, NULL);

  png_destroy_read_struct (&png_ptr, &info_ptr, &end_info);

  fclose (fp);

  *Rw = ww;
  *Rh = hh;

  return 0;

err_ret_destroy0:
  png_destroy_read_struct (&png_ptr, 0, 0);
  fclose (fp);
  return errstr;
err_ret_destroy1:
  png_destroy_read_struct (&png_ptr, &info_ptr, 0);
  fclose (fp);
  return errstr;
err_ret_destroy2:
  png_destroy_read_struct (&png_ptr, &info_ptr, &end_info);
  fclose (fp);
  return errstr;
err_ret_close:
  fclose (fp);
  return errstr;
}

static char *do_read_graymap
  (png_structp png_ptr,
   unsigned int *Rw, unsigned int *Rh, unsigned char **Rdata)
{
  char *errstr;
  png_infop info_ptr;
  png_infop end_info;
  unsigned char *data, **rows;
  png_uint_32 ww, hh, i;
  int bit_depth, color_type, interlace_type;

  info_ptr= NULL;
  end_info= NULL;
  rows= NULL;
  data= NULL;

  info_ptr = png_create_info_struct (png_ptr);
  if (!info_ptr)
  {
    errstr= strdup("internal error. libpng: "
                    "can not allocate first info struct.");
    goto done;
  }

  end_info = png_create_info_struct (png_ptr);
  if (!end_info)
  {
    errstr= strdup("internal error. libpng: "
                    "can not allocate second info struct.");
    goto done;
  }

  if (setjmp (png_jmpbuf (png_ptr)))
  {
    errstr= strdup("internal error. libpng jumped back.");
    goto done;
  }

  // Get all of the information from the PNG file before the first IDAT (imaged ata chunk).
  png_read_info (png_ptr, info_ptr);

  // Get the goods
  png_get_IHDR (png_ptr, info_ptr, &ww, &hh, &bit_depth, &color_type,
		&interlace_type, NULL, NULL);

  // tell libpng to strip 16 bit/color files down to 8 bits/color
  png_set_strip_16 (png_ptr);

  // Expand paletted colors into true RGB triplets
  if (color_type == PNG_COLOR_TYPE_PALETTE)
    png_set_expand (png_ptr);

  // Expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel
  if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
    png_set_expand (png_ptr);

#if 0
  // Expand paletted or RGB images with transparency to full alpha channels
  // so the data will be available as RGBA quartets.
  if (png_get_valid (png_ptr, info_ptr, PNG_INFO_tRNS))
    png_set_expand (png_ptr);
#endif

  if (color_type!=PNG_COLOR_TYPE_GRAY && color_type!=PNG_COLOR_TYPE_GRAY_ALPHA)
     png_set_rgb_to_gray_fixed(png_ptr, 1, -1, -1);

  if (color_type & PNG_COLOR_MASK_ALPHA)
        png_set_strip_alpha(png_ptr);

  // Turn on interlace handling
  png_set_interlace_handling (png_ptr);

  // Update image info based on transformations
  png_read_update_info (png_ptr, info_ptr);

  data = malloc (ww * hh * sizeof (unsigned char));
  rows = malloc (hh * sizeof (unsigned char *));
  for (i = 0; i < hh; i++)
    rows[i] = data + i * ww;

  png_read_image (png_ptr, (void *) rows);

  // read rest of file, and get additional chunks in info_ptr
  png_read_end (png_ptr, info_ptr);

  // clean up after the read, and free any memory allocated
  png_destroy_read_struct (&png_ptr, &info_ptr, &end_info);

  free(rows);

  *Rw = ww;
  *Rh = hh;
  *Rdata = data;

  return NULL;
done:
  if (data) free(data);
  if (rows) free(rows);
  png_destroy_read_struct (&png_ptr, info_ptr ? &info_ptr : NULL,
                                     end_info ? &end_info : NULL);
  return errstr;
}

char *read_png_graymap_noalpha
  (const char *fn, unsigned int *Rw, unsigned int *Rh, unsigned char **Rdata)
{
  char *errstr;
  FILE *fp;
  png_structp png_ptr;

  if ((errstr=open_png_file(fn,&fp))) return errstr;

  png_ptr= png_create_read_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (!png_ptr)
  {
    errstr= strdup ("internal error. libpng: can not allocate PNG struct.");
      goto done;
  }

  png_init_io (png_ptr, fp);
  png_set_sig_bytes (png_ptr, SIG_BYTES);

  errstr= do_read_graymap(png_ptr, Rw, Rh, Rdata);
done:
  fclose(fp);
  return errstr;
}

char *read_png_g8_filep
  (FILE *fp, unsigned int *Rw, unsigned int *Rh, unsigned char **Rdata,int *eof)
{
  png_structp png_ptr;
  unsigned char header[SIG_BYTES];

  if (fread(header, 1, 1, fp)!=1)
  {
    if (feof(fp)) *eof= 1;
             else *eof= 0;
    return strdup("can not read signature bytes from stream");
  }

  if (fread(header+1, 1, SIG_BYTES-1, fp)!=SIG_BYTES-1)
    return strdup("can not read signature bytes from stream");

  if (png_sig_cmp (header, 0, SIG_BYTES))
    return strdup("stream doesn't match PNG magic");

  png_ptr= png_create_read_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (!png_ptr)
    return strdup ("internal error: libpng can not allocate PNG struct.");

  png_init_io (png_ptr, fp);
  png_set_sig_bytes (png_ptr, SIG_BYTES);

  return do_read_graymap(png_ptr, Rw, Rh, Rdata);
}

typedef struct ibufnode
{
  unsigned int len;
  unsigned char buffer[1000];
  struct ibufnode *next;
} ibufnode_t;

typedef struct 
{
  ibufnode_t *first;
  ibufnode_t *last;
  unsigned int total;
} ibuf_t;

static void ibuf_destroy(ibuf_t *ib)
{
  ibufnode_t *B,*N;
  for(B=ib->first;B;B=N)
  {
    N= B->next;
    free(B);
  }
}

static void ibuf_write(ibuf_t *ib, unsigned char *data, unsigned int len)
{
  unsigned int cp;
  while(len)
  {
    cp= sizeof(ib->last->buffer) - ib->last->len;
    if (cp==0)
    {
      ibufnode_t *B;
      B= malloc(sizeof(*B));
      B->next= NULL;
      B->len= 0;
      ib->last->next= B;
      ib->last= B;
      continue; 
    }
    if (cp>len) cp= len;
    memcpy(ib->last->buffer + ib->last->len, data, cp);
    ib->last->len+= cp;
    len-= cp;
    data+= cp;
    ib->total+= cp;
  }
}

static void* ibuf_collectf(ibuf_t *ib, unsigned int *Rl)
{
  char* d;
  ibufnode_t *B,*N;
  unsigned int off;

  d= malloc(ib->total);
  off= 0;
  for(B=ib->first;B;B=N)
  {
    N= B->next;
    memcpy(d+off, B->buffer, B->len);
    off+= B->len;
    free(B);
  }
  *Rl= ib->total;
  return d;
}

static void ibuf_out(png_structp png_ptr, png_bytep data, png_uint_32 len)
{
  ibuf_write( png_get_io_ptr(png_ptr), data, len );
}

static void dummy_flush(png_structp png_ptr)
{
}


char *encode_png_g8
  (unsigned int w, unsigned int h, unsigned char *data,
   unsigned int *Rlen, unsigned char **Rdata)
{
  char *errstr;
  char ebu[2000];
  png_structp png_ptr;
  png_infop info_ptr;
  uint8_t **rows;
  png_uint_32 i;
  ibuf_t buffer;

  rows= NULL; 
  errstr= NULL;

  buffer.first= buffer.last= malloc(sizeof(buffer.first[0]));
  buffer.first->next= NULL;
  buffer.first->len= 0;
  buffer.total= 0;
  

  png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (!png_ptr)
  {
    errstr= strdup ("internal error. libpng: can not allocate PNG struct.");
    goto err_ret_close;
  }

  info_ptr = png_create_info_struct (png_ptr);
  if (!info_ptr)
  {
    errstr= strdup("internal error. libpng: can not allocate info struct.");
    goto err_ret_destroy0;
  }

  if (setjmp(png_jmpbuf (png_ptr)))
  {
    errstr = strdup ("internal error. libpng jumped back.");
    goto err_ret_destroy1;
  }

  png_set_write_fn(png_ptr, &buffer,
                   (png_rw_ptr) ibuf_out,
                   (png_flush_ptr)dummy_flush);

  png_set_IHDR (png_ptr, info_ptr,
		w, h, 8,
		PNG_COLOR_TYPE_GRAY,
		PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

  png_write_info (png_ptr, info_ptr);

  rows = malloc(h*sizeof(unsigned char *));
  for(i=0;i<h;i++) rows[i] = data + i*w;

  png_write_image (png_ptr, (void *) rows);
  png_write_end (png_ptr, info_ptr);

  png_destroy_write_struct (&png_ptr, &info_ptr);
  free (rows);

  *Rdata= ibuf_collectf(&buffer, Rlen);

  return 0;

err_ret_destroy0:
  ibuf_destroy(&buffer);
  png_destroy_write_struct(&png_ptr, NULL);
  return errstr;
err_ret_destroy1:
  ibuf_destroy(&buffer);
  if (rows) free(rows);
  png_destroy_write_struct(&png_ptr, &info_ptr);
  return errstr;
err_ret_close:
  ibuf_destroy(&buffer);
  return errstr;
}

char *decode_png_g8
  (const unsigned char *data, size_t len,
   unsigned int *Rw, unsigned int *Rh, unsigned char **Rdata)
{
  pbuf_t buffer;
  png_structp png_ptr;

  if (png_sig_cmp (data, 0, SIG_BYTES))
    return strdup("Not a valid PNG buffer.");

  png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);

  if (!png_ptr)
    return strdup("internal error. libpng: can not allocate PNG struct.");

  buffer.start= data+SIG_BYTES; buffer.end= data+len;
  png_set_read_fn(png_ptr, &buffer, read_buffer);
  png_set_sig_bytes (png_ptr, SIG_BYTES);

  return do_read_graymap(png_ptr, Rw, Rh, Rdata);
}

char *encode_png_argb
  (unsigned int w, unsigned int h, unsigned int *data,
   unsigned int *Rlen, unsigned char **Rdata)
{
  char *errstr;
  char ebu[2000];
  png_structp png_ptr;
  png_infop info_ptr;
  uint8_t **rows;
  png_uint_32 i;
  ibuf_t buffer;
  uint8_t *tdata;

  rows= NULL; 
  errstr= NULL;
  tdata= NULL;

  buffer.first= buffer.last= malloc(sizeof(buffer.first[0]));
  buffer.first->next= NULL;
  buffer.first->len= 0;
  buffer.total= 0;
  

  png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (!png_ptr)
  {
    errstr= strdup ("internal error. libpng: can not allocate PNG struct.");
    goto err_ret_close;
  }

  info_ptr = png_create_info_struct (png_ptr);
  if (!info_ptr)
  {
    errstr= strdup("internal error. libpng: can not allocate info struct.");
    goto err_ret_destroy0;
  }

  if (setjmp(png_jmpbuf (png_ptr)))
  {
    errstr = strdup ("internal error. libpng jumped back.");
    goto err_ret_destroy1;
  }

  png_set_write_fn(png_ptr, &buffer,
                   (png_rw_ptr) ibuf_out,
                   (png_flush_ptr)dummy_flush);

  png_set_IHDR (png_ptr, info_ptr,
		w, h, 8,
		PNG_COLOR_TYPE_RGBA,
		PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

  png_write_info (png_ptr, info_ptr);

  rows = malloc (h * sizeof (unsigned int *));
  tdata= malloc(sizeof(uint8_t)*w*h*4);
  for(i=0;i<w*h;i++)
  {
    tdata[i*4+0]= (data[i] >> 16);   // R
    tdata[i*4+1]= (data[i] >> 8);    // G
    tdata[i*4+2]= (data[i] >> 0);    // B
    tdata[i*4+3]= (data[i] >> 24);   // A
  }
  for(i=0;i<h;i++)
    rows[i] = tdata + i*w*4;

  png_write_image (png_ptr, (void *) rows);
  png_write_end (png_ptr, info_ptr);

  png_destroy_write_struct (&png_ptr, &info_ptr);
  free(tdata);
  free (rows);

  *Rdata= ibuf_collectf(&buffer, Rlen);

  return 0;

err_ret_destroy0:
  ibuf_destroy(&buffer);
  png_destroy_write_struct(&png_ptr, NULL);
  return errstr;
err_ret_destroy1:
  ibuf_destroy(&buffer);
  if (rows) free(rows);
  if (tdata) free(tdata);
  png_destroy_write_struct(&png_ptr, &info_ptr);
  return errstr;
err_ret_close:
  ibuf_destroy(&buffer);
  return errstr;
}

char *encode_png_rgb
  (unsigned int w, unsigned int h, unsigned int *data,
   unsigned int *Rlen, unsigned char **Rdata)
{
  char *errstr;
  char ebu[2000];
  png_structp png_ptr;
  png_infop info_ptr;
  uint8_t **rows;
  png_uint_32 i;
  ibuf_t buffer;
  uint8_t *tdata;

  rows= NULL; 
  errstr= NULL;
  tdata= NULL;

  buffer.first= buffer.last= malloc(sizeof(buffer.first[0]));
  buffer.first->next= NULL;
  buffer.first->len= 0;
  buffer.total= 0;
  

  png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (!png_ptr)
  {
    errstr= strdup ("internal error. libpng: can not allocate PNG struct.");
    goto err_ret_close;
  }

  info_ptr = png_create_info_struct (png_ptr);
  if (!info_ptr)
  {
    errstr= strdup("internal error. libpng: can not allocate info struct.");
    goto err_ret_destroy0;
  }

  if (setjmp(png_jmpbuf (png_ptr)))
  {
    errstr = strdup ("internal error. libpng jumped back.");
    goto err_ret_destroy1;
  }

  png_set_write_fn(png_ptr, &buffer,
                   (png_rw_ptr) ibuf_out,
                   (png_flush_ptr)dummy_flush);

  png_set_IHDR (png_ptr, info_ptr,
		w, h, 8,
		PNG_COLOR_TYPE_RGB,
		PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

  png_write_info (png_ptr, info_ptr);

  rows = malloc (h * sizeof (unsigned int *));
  tdata= malloc(sizeof(uint8_t)*w*h*3);
  for(i=0;i<w*h;i++)
  {
    tdata[i*3+0]= (data[i] >> 16);   // R
    tdata[i*3+1]= (data[i] >> 8);    // G
    tdata[i*3+2]= (data[i] >> 0);    // B
  }
  for(i=0;i<h;i++)
    rows[i] = tdata + i*w*3;

  png_write_image (png_ptr, (void *) rows);
  png_write_end (png_ptr, info_ptr);

  png_destroy_write_struct (&png_ptr, &info_ptr);
  free(tdata);
  free (rows);

  *Rdata= ibuf_collectf(&buffer, Rlen);

  return 0;

err_ret_destroy0:
  ibuf_destroy(&buffer);
  png_destroy_write_struct(&png_ptr, NULL);
  return errstr;
err_ret_destroy1:
  ibuf_destroy(&buffer);
  if (rows) free(rows);
  if (tdata) free(tdata);
  png_destroy_write_struct(&png_ptr, &info_ptr);
  return errstr;
err_ret_close:
  ibuf_destroy(&buffer);
  return errstr;
}

char *write_png_g8
  (const char *fn, unsigned int w, unsigned int h, unsigned char *data)
{
  char *errstr;
  FILE *fp;
  char ebu[2000];
  png_structp png_ptr;
  png_infop info_ptr;
  uint8_t **rows;
  png_uint_32 i;

  rows= NULL; 

  errstr = 0;
  fp = fopen (fn, "wb");
  if (!fp)
  {
    snprintf(ebu,sizeof(ebu),"can not open file %s: %s", fn, strerror (errno));
    return strdup(ebu);
  }

  png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (!png_ptr)
  {
    errstr= strdup ("internal error. libpng: can not allocate PNG struct.");
    goto err_ret_close;
  }

  info_ptr = png_create_info_struct (png_ptr);
  if (!info_ptr)
  {
    errstr= strdup ("internal error. libpng: can not allocate info struct.");
    goto err_ret_destroy0;
  }

  if (setjmp (png_jmpbuf (png_ptr)))
  {
    errstr = strdup ("internal error. libpng jumped back.");
    goto err_ret_destroy1;
  }

  png_init_io (png_ptr, fp);

  png_set_IHDR (png_ptr, info_ptr,
		w, h, 8,
		PNG_COLOR_TYPE_GRAY,
		PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

  png_write_info (png_ptr, info_ptr);

  rows = malloc(h*sizeof(unsigned char *));
  for(i=0;i<h;i++) rows[i] = data + i*w;

  png_write_image (png_ptr, (void *) rows);
  png_write_end (png_ptr, info_ptr);

  png_destroy_write_struct (&png_ptr, &info_ptr);

  free (rows);
  fclose (fp);

  return 0;

err_ret_destroy0:
  png_destroy_write_struct (&png_ptr, 0);
  fclose (fp);
  return errstr;
err_ret_destroy1:
  if (rows) free(rows);
  png_destroy_write_struct (&png_ptr, &info_ptr);
  fclose (fp);
  return errstr;
err_ret_close:
  fclose (fp);
  return errstr;
}

